// cbm_virtual_hd.c
// Emulation of a Commodore PET compatible disk drive using a standard
// GPIB card under Linux.
// By Tim Howe, moc.dlrowltn@321ylbmit February 2010
// Released under BSD licence
//
// Thanks to the original author(s) of the Non-controller-in-charge
// example on the National Instruments Website for a crucial insight
// into how to write such an application using the NI-488.2 API
//
// Thanks also to the Linux GPIB team for writing and maintaining
// the driver and API libeary for Linux
//
// Thanks also to my son Ben for inspiring me to finish something
// I planned to do way back in 2002!
//
// Still to do:
//   Finish the function that generates the directory listing
//   Perhaps add some timeouts to the FSM to make it more robust
//

#define PROG_VERSION    "V0.80"

#ifdef HAVE_CONFIG_H
    #include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <termios.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include "gpib/ib.h"

#define _DEBUG

// Debugging
#ifdef _DEBUG
   #define dbgPrint(data)  printf data
#else
   #define dbgPrint(data)
#endif

// Defines
#define PROG_BUFFER_LENGTH  32768
#define SLEEP_TIME          50
#define GPIB_BOARD          0
#define GPIB_ADDR           8
#define MAX_FILENAME        64
#define MAX_PATH            1024

// Types
typedef enum
{
    tIdle,
    tRead,
    tWrite

} tIoType;

// Function Prototypes
int InitGpib(int *board, int pri_addr);
void DeviceLoop (int bid, char *InBuffer, char *OutBuffer, char *);
int load_program(char *filename, char *path, char *buffer, int max_len);
int save_program(char *filename, char *path, char *buffer, int len);
int create_dir_listing(char *buffer);
int keypress (void);



// Program Entry Point
int main(int argc, char *argv[])
{
    int board_id = GPIB_BOARD;
    char* InBuffer;
    char* OutBuffer;
    char path[MAX_PATH];

    // Process command line arguments
    if( argc < 2 )
    {
        printf("\nVirtual IEEE488 hard disk for CBM PET using a standard GPIB card under Linux.\n");
        printf("%s (C) Tim Howe, Feb 2010, moc.dlrowltn@321ylbmit\n", PROG_VERSION);
        printf("Usage: cbm_virtual_hd <path to program files>\n\n");
        return -1;
    }


    // Initialise the GPIB card
    if (InitGpib(&board_id, GPIB_ADDR)  < 0)
    {
        return -1;
    }

    // Allocate in and out buffers
    InBuffer = malloc(PROG_BUFFER_LENGTH);
    if (InBuffer == NULL)
    {
        printf("Unable to allocate memory\n");
        //ibonl (board_id, 0);
        return -1;
    }

    OutBuffer = malloc(PROG_BUFFER_LENGTH);
    if (OutBuffer == NULL)
    {
        printf("Unable to allocate memory\n");
        free(InBuffer);
        //ibonl (board_id, 0);
        return -1;
    }


    printf("CBM Virtual Hard Disk running on GPIB0 device id %d\n", GPIB_ADDR);
    strncpy(path, argv[1], MAX_PATH-MAX_FILENAME-2);
    if(path[strlen(path)-1] != '/') strcat(path, "/");
    printf("HD directory set to: %s\n", path);

    // Begin Device loop
    DeviceLoop (board_id, InBuffer, OutBuffer, path);


    // Turn board_id offline
    //ibonl (board_id, 0);

    // Delete the message Buffers
    free(InBuffer);
    free(OutBuffer);

    return 0;
}



// Initialise the GPIB board.
// (Note that most of the clobal board configuration has been
// specified in /etc/gpib.conf

int InitGpib(int *board, int pri_addr)
{
#if 0
    // Open a handle to the board by name
    *board = ibfind ("GPIB0");
    if (ibsta & ERR)
    {
        printf("Unable to find the board, iberr = %d\n", iberr);
        return -1;
    }

    // Disable System Control capability
    ibrsc (*board, 0);
    if (ibsta & ERR)
    {
        printf("Unable to relinquish System Control capability\n");
        ibonl (*board, 0);
        return -1;
    }
#endif

    // Set Primary address
    ibpad (*board, pri_addr);
    if (ibsta & ERR)
    {
        printf("Unable to change Primary Address\n");
        ibonl (*board, 0);
        return -1;
    }

    // Set timeout to 10s
    // (A 32k byte transfer will take ~6.5s at 5000bytes/s)
    ibtmo(*board, T10s);
    if( ibsta & ERR )
    {
        printf("Unable to set timeout to 10 sec\n");
        ibonl (*board, 0);
        return -1;
    }

    // Ensure EOS by character is turned off
    ibeos(*board, 0);
    if( ibsta & ERR )
    {
        printf("Unable to disable EOS character\n");
        ibonl (*board, 0);
        return -1;
    }

    return 0;
}



// This is the main device loop. We perform all I/O using asynchronous calls.
// This prevents any timeout errors and also prevents us from delaying the
// system if we make two writes or two reads in a row.
//
// We do an ibwait 0 to update the status, and then we look at the status
// bits to see if any of the bits that we care about are set. Since we
// care about different bits at different times of the device's life, we
// have a mask which we update and compare against. This device understands
// Device Clear, as well as being capable of talking and listening.
//
// The order of checking the bits is somewhat important. We must first check
// for Device Clear. The other bits can be checked in any order. When we
// receive a Device Clear, we will clear out the buffers and return back
// to the IDLE state.
//
// When we become a listener, we first check to see if we have an asynchronous
// write outstanding. If we do, we abort it. Then, we start an asynchronous
// read.
//
// When we become a talker, we first check to see if we have an asynchronous
// read outstanding. If we do, we abort it. Then, we start an asynchronous
// write. Since we may not have anything to say, we only begin the write if
// we have something to say.
//
// Finally, if the Asynchronous I/O finishes, we check to see what type of
// I/O was in use. If we were performing a read, we quickly run the input
// buffer through a parser and act accordingly. If we were performing a
// write, we just note that we finished a write.

void DeviceLoop (int bid, char *InBuffer, char *OutBuffer, char *file_path)
{
    int mask, i;
    tIoType IoType;
    char filename[MAX_FILENAME];
    int file_len=0;



    printf("Entering loop, press any key to exit.\n");


    // Clear all the buffers
    memset(InBuffer, 0, PROG_BUFFER_LENGTH);
    memset(OutBuffer, 0, PROG_BUFFER_LENGTH); file_len = 0;
    memset(filename, 0, MAX_FILENAME);

    // When we are idle, we only care about being Cleared or being addressed
    // to be a Talker or a Listener. We use a mask to differentiate between
    // conditions that appear on the bus and conditions that we care about on
    // the bus.
    IoType = tIdle;
    mask = (DCAS | TACS | LACS);

   // Just loop forever for now. Normally devices just quit when they get
   // powered off. We do have a check and if the user presses a key on the
   // keyboard, we will exit the loop and the program.
   while (1)
   {
        // If the user presses the keyboard, exit.
        //if (_kbhit())
        if ( keypress() >= 0 )
        {
            printf("Key pressed: Quitting cbm_virtual_hd.\n");
            return;
        }


        // Update status (non-blocking)
        ibwait (bid, 0);

        // We received a Device Clear
        if (ibsta & mask & DCAS)
        {
            dbgPrint(("Device Clear Received\n"));

            // Check to see if DCAS is still in the hardware. If DCAS terminated
            // I/O, it will already be cleared, If I/O was idle, DCAS needs to
            // be cleared from the hardware manually
            ibwait (bid, 0);
            if (ibsta & DCAS)
            {
                // Clear DCAS from the Hardware Manually
                ibwait (bid, DCAS);
            }

            // Clear Everything.
            memset(InBuffer, 0, PROG_BUFFER_LENGTH);
            memset(OutBuffer, 0, PROG_BUFFER_LENGTH); file_len = 0;
            memset(filename, 0, MAX_FILENAME);

            // Since we are going back to IDLE, again, check for Device Clear
            // or addressed to Talk to Listen.
            IoType = tIdle;
            mask = (DCAS | TACS | LACS);
        }

        // We are addressed as a Listener
        else if (ibsta & mask & LACS)
        {
            dbgPrint(("Addressed as a Listener\n"));

            // If a write is in progress, cancel it
            if (IoType == tWrite)
            {
                dbgPrint(("   Cancelling Write\n"));
                ibstop (bid);
            }

            // Start a read
            dbgPrint(("   Starting Read\n"));
            ibrda (bid, InBuffer, PROG_BUFFER_LENGTH);

            // Since we have started a read, we care if we are cleared,
            // get addressed to Talk, or if the I/O (read) completes.
            IoType = tRead;
            mask = (DCAS | TACS | CMPL);
        }

        // We are addressed as a Talker
        else if (ibsta & mask & TACS)
        {
            dbgPrint(("Addressed as a Talker\n"));

            // If a read is in progress, cancel it
            if (IoType == tRead)
            {
                dbgPrint(("   Cancelling Read\n"));
                ibstop (bid);
            }

            // If we already have a filename then assume this is a LOAD operation
            if (strlen (filename) > 0)
            {
                if (!strncmp(filename, "$", 1))
                {
                    printf("Sending a directory listing\n");
                    file_len = create_dir_listing(OutBuffer);
                }
                else
                {

                    file_len = load_program(filename, file_path, OutBuffer, PROG_BUFFER_LENGTH);
                    if (file_len > 0)
                    {
                        printf("Loaded '%s' from disk (%d bytes)\n", filename, file_len);
                    }
                    else
                    {
                        printf("Could not load file '%s' from disk\n", filename);
                        memset(OutBuffer, 0, PROG_BUFFER_LENGTH); file_len = 0;
                    }
                }
            }

            // Start a write if we have data to send
            if (file_len > 0)
            {
                dbgPrint(("   Starting Write\n"));
                //ibwrta (bid, OutBuffer, strlen(OutBuffer));
                ibwrta (bid, OutBuffer, file_len);

                // Since we have started a write, we care if we are cleared,
                // get addressed to Listen, or if the I/O (read) completes.
                IoType = tWrite;
                mask = (DCAS | LACS | CMPL);
            }
            else
            {
                dbgPrint(("   Nothing to Say\n"));

                // Clear Everything.
                memset(InBuffer, 0, PROG_BUFFER_LENGTH);
                //memset(OutBuffer, 0, PROG_BUFFER_LENGTH); file_len = 0;
                memset(filename, 0, MAX_FILENAME);

                // Since we have nothing to say, we do not care if we are
                // addressed as a talker (since we will continue to have
                // nothing to say). Therefore, we care if we are cleared,
                // or get addressed to Listen.
                IoType = tIdle;
                mask = (DCAS | LACS); // TDH: Interesting no TACS flag here
            }
        }

        // I/O Completed
        else if (ibsta & mask & CMPL)
        {
            printf("I/O Complete: %d bytes transferred\n", ThreadIbcnt());

            // If a read is complete, wait again for DCAS, TACS and LACS
            if (IoType == tRead)
            {
                // If we do not already have a filename then take this as the filename
                if (strlen (filename) == 0)
                {
                    printf("Taking string '%s' as filename\n", InBuffer);

                    // Copy filename and converting to lower case
                    for (i=0; i<MAX_FILENAME; i++)
                    {
                        if (!InBuffer[i]) break;
                        filename[i] = tolower(InBuffer[i]);
                    }
                }

                else // Assume that this is the data part of a SAVE operation
                {
                    file_len = save_program(filename, file_path, InBuffer, ThreadIbcnt());
                    if (file_len > 0)
                    {
                        printf("Saved %d bytes to file '%s'\n", ThreadIbcnt(), filename);
                    }
                    else
                    {
                        printf("Unable to save file '%s'\n", filename);
                    }

                    memset(InBuffer, 0, PROG_BUFFER_LENGTH);
                    memset(OutBuffer, 0, PROG_BUFFER_LENGTH); file_len = 0;
                    memset(filename, 0, MAX_FILENAME);
                }
            }

            // If a write is complete, just wait for DCAS and LACS, since
            // even if we are told to write again, we have nothing to say.
            else if (IoType == tWrite)
            {
                // We just finished the write. Zero-out the output buffer
                // since we have nothing more to say now.
                //printf("GPIB status ibcnt: %d bytes sent\n", ThreadIbcnt());
                memset(InBuffer, 0, PROG_BUFFER_LENGTH);
                memset(OutBuffer, 0, PROG_BUFFER_LENGTH); file_len = 0;
                memset(filename, 0, MAX_FILENAME);
            }

            // Since the I/O is complete, we are going back to IDLE. Check
            // for Device Clear or addressed to Talk to Listen.
            IoType = tIdle;
            mask = (DCAS | TACS | LACS);

        }


        // Otherwise, we just did an ibwait 0 and we did not detect any wait
        // bits that we care about. Just sleep a bit to allow the uP to process
        // other events. We could also do an ibwait TIMO and set the timeout to
        // be a small value. Either way works.
        else
        {
            usleep(SLEEP_TIME);
        }

   } // end while

}



// Open the specified file and determine how big it is.
// If the file exists, read it into the supplied buffer
// otherwise return -1 indicating an error.

int load_program(char *filename, char *path, char *buffer, int max_len)
{
    FILE *fd;
    struct stat fst;
    char abs_filename[MAX_PATH];

    strncpy(abs_filename, path, MAX_PATH-MAX_FILENAME);
    strncat(abs_filename, filename, MAX_PATH);

    stat(abs_filename, &fst);
    fd = fopen(abs_filename, "rb");
    if (fd != NULL && fst.st_size < max_len)
    {
        fread(buffer, fst.st_size, 1, fd);
        fclose(fd);
        return (fst.st_size);
    }
    else return (-1);
}



// Save file & return number of bytes written or -1 for error
// (This will attempt not to overwrite an existing file.)

int save_program(char *filename, char *path, char *buffer, int len)
{
    FILE *fd;
    char abs_filename[MAX_PATH];

    strncpy(abs_filename, path, MAX_PATH-MAX_FILENAME);
    strncat(abs_filename, filename, MAX_PATH);

    // Test if file already exists
    fd = fopen(abs_filename, "rb");
    if (fd != NULL)
    {
        printf("File %s already exists: ", filename);
        fclose(fd);
        return (-1);
    }

    fd = fopen(abs_filename, "wb");
    if (fd != NULL)
    {
        len = fwrite(buffer, len, 1, fd);
        fclose(fd);
        return (len);
    }
    else return (-1);
}



// Create a directory listing to send to the PET
// This function is just a placeholder for now

char dummy[] =
{
    0x01, 0x04, 0x09, 0x04, 0x0a, 0x00, 0x44, 0x49,
    0x52, 0x00, 0x00, 0x00
};

int create_dir_listing(char *buffer)
{
    memcpy(buffer, dummy, 12);
    return 12;
}


// Checks keyboard buffer (stdin) and returns key
// pressed, or -1 for no key pressed

int keypress (void)
{
    static char keypressed;
    struct timeval waittime;
    int num_chars_read;
    fd_set mask;
    FD_SET(0, &mask);

    waittime.tv_sec = 0;
    waittime.tv_usec = 0;
    if (select (1, &mask, 0, 0, &waittime))
    {
        num_chars_read = read (0, &keypressed, 1);
        if (num_chars_read == 1)
            return ((int)keypressed);
    }

    return (-1);
}


